//	Altirra - Atari 800/800XL/5200 emulator
//	Copyright (C) 2009-2022 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License along
//	with this program. If not, see <http://www.gnu.org/licenses/>.

#include "stdafx.h"
#include <windows.h>
#include <windowsx.h>
#include <vd2/system/binary.h>
#include <vd2/system/bitmath.h>
#include <vd2/system/color.h>
#include <vd2/system/cpuaccel.h>
#include <vd2/system/file.h>
#include <vd2/system/vdstl_vectorview.h>
#include <vd2/system/zip.h>
#include <vd2/Dita/services.h>
#include <at/atcore/atascii.h>
#include <at/atnativeui/canvas_win32.h>
#include <at/atnativeui/uinativewindow.h>
#include "oshelper.h"
#include "printeroutput.h"
#include "printerttfencoder.h"
#include "resource.h"
#include "simulator.h"
#include "texteditor.h"
#include "uidbgprinteroutput.h"

extern ATSimulator g_sim;

///////////////////////////////////////////////////////////////////////////////

class ATUIPrinterGraphicalOutputWindow final : public ATUINativeWindow {
public:
	ATUIPrinterGraphicalOutputWindow();

	void AttachToOutput(ATPrinterGraphicalOutput& output);

	void Clear();

	void ResetView();
	void SaveAsPNG(float dpi);
	void SaveAsPDF();

private:
	LRESULT WndProc(UINT msg, WPARAM wParam, LPARAM lParam) override;
	LRESULT OnNCCalcSize(WPARAM wParam, LPARAM lParam);
	void OnDestroy();
	void OnSize();
	void OnKeyDown(uint32 vk, uint32 flags);
	void OnMouseMove(int x, int y);
	void OnMouseDownL(int x, int y);
	void OnMouseUpL(int x, int y);
	void OnMouseWheel(int x, int y, float delta);
	void OnCaptureChanged();
	bool OnGesture(WPARAM wParam, LPARAM lParam);
	void OnPaint();
	void OnDpiChanged(int dpi);

	void UpdateViewOrigin();
	void ScrollByPixels(int dx, int dy);
	void SetZoom(float clicks);
	void SetZoom(float clicks, const vdpoint32& centerPt);
	void UpdateZoom(const vdpoint32 *centerPt);

	void ForceFullInvalidation();
	void ProcessPendingInvalidation();
	void InvalidatePaperArea(float x1, float y1, float x2, float y2);

	vdrect32f GetDocumentBounds() const;

	struct ViewTransform {
		float mOriginX = 0;
		float mOriginY = 0;
		float mPixelsPerMM = 1;
		float mMMPerPixel = 1;
	};

	bool Render(const ViewTransform& viewTransform, sint32 x, sint32 y, uint32 w, uint32 h, bool force);
	void Downsample8x8(uint32 *dst, const uint8 *src, size_t w);
	void Downsample8x8_Scalar(uint32 *dst, const uint8 *src, size_t w);
#if defined(VD_CPU_X64)
	void Downsample8x8_POPCNT64(uint32 *dst, const uint8 *src, size_t w);
#elif defined(VD_CPU_X86)
	void Downsample8x8_POPCNT32(uint32 *dst, const uint8 *src, size_t w);
#endif

	float mPageWidthMM = 0;
	float mPageVBorderMM = 0;
	float mDotSizeMM = 0;

	float mViewOriginX = 0;
	float mViewOriginY = 0;
	float mViewCenterX = 0;
	float mViewCenterY = 0;
	float mViewPixelsPerMM = 1.0f;
	float mViewMMPerPixel = 1.0f;
	sint32 mViewWidthPixels = 0;
	sint32 mViewHeightPixels = 0;
	int mViewDpi = 96;

	bool mbDragging = false;
	sint32 mDragLastX = 0;
	sint32 mDragLastY = 0;
	float mWheelAccum = 0;

	bool mbInGesture = false;
	bool mbFirstGestureEvent = false;
	vdpoint32 mGestureOrigin;
	float mGestureZoomOrigin = 0;

	static constexpr float kZoomMin = -10.0f;
	static constexpr float kZoomMax = 25.0f;
	float mZoomClicks = 0;

	bool mbInvalidationPending = false;

	using RenderDot = ATPrinterGraphicalOutput::RenderDot;
	vdfastvector<RenderDot> mDotCullBuffer;

	ATPrinterGraphicalOutput *mpOutput = nullptr;
	ATUICanvasW32 mCanvas;
	vdfastvector<uint32> mFrameBuffer;
	vdfastvector<uint8> mABuffer;

	vdstructex<RGNDATA> mRegionData;

	uint8 mPopCnt8[256];
	uint32 mGammaTable[65];

	static constexpr float kBlackLevel = 0.03f;
	static constexpr float kWhiteLevel = 0.90f;
};

ATUIPrinterGraphicalOutputWindow::ATUIPrinterGraphicalOutputWindow() {
	for(int i=0; i<256; ++i) {
		uint8 v = i;

		v -= (v >> 1) & 0x55;
		v = (v & 0x33) + ((v & 0xCC) >> 2);

		mPopCnt8[i] = (v & 0x0F) + (v >> 4);
	}

	static constexpr float kRange = kWhiteLevel - kBlackLevel;
	for(int i=0; i<=64; ++i) {
		float luma = kWhiteLevel - kRange * (float)i / 64.0f;
		uint32 v = VDColorRGB(vdfloat32x3::set(luma, luma, luma)).LinearToSRGB().ToRGB8();

		mGammaTable[i] = v;
	}
}

void ATUIPrinterGraphicalOutputWindow::AttachToOutput(ATPrinterGraphicalOutput& output) {
	const ATPrinterGraphicsSpec& spec = output.GetGraphicsSpec();

	mpOutput = &output;
	mpOutput->SetOnInvalidation(
		[this] {
			if (!mbInvalidationPending) {
				mbInvalidationPending = true;

				::PostMessage(mhwnd, WM_USER + 100, 0, 0);
			}
		}
	);

	mPageWidthMM = spec.mPageWidthMM;
	mPageVBorderMM = spec.mPageVBorderMM;
	mDotSizeMM = spec.mDotSizeMM;

	ResetView();
}

void ATUIPrinterGraphicalOutputWindow::Clear() {
	mViewCenterY = 0;

	UpdateViewOrigin();

	if (mpOutput)
		mpOutput->Clear();
}

void ATUIPrinterGraphicalOutputWindow::ResetView() {
	mZoomClicks = -1000.0f;
	SetZoom(0);

	mViewCenterX = mPageWidthMM * 0.5f;
	mViewCenterY = mPageVBorderMM;

	UpdateViewOrigin();

	ForceFullInvalidation();
}

void ATUIPrinterGraphicalOutputWindow::SaveAsPNG(float dpi) {
	const VDStringW& fn = VDGetSaveFileName("PrinterSaveAsPNG"_vdtypeid, (VDGUIHandle)mhwnd, L"Save PNG image", L"PNG image\0*.png\0", L"png");
	if (fn.empty())
		return;

	// get bounds in document space
	vdrect32f documentBounds = GetDocumentBounds();

	// compute pixel bounds
	const float mmToInches = 1.0f / 25.4f;
	const sint32 w = std::max<sint32>(1, (sint32)ceilf(documentBounds.width() * mmToInches * dpi));
	const sint32 h = std::max<sint32>(1, (sint32)ceilf(documentBounds.height() * mmToInches * dpi));

	// render whole screen
	ViewTransform vt;
	vt.mOriginX = documentBounds.left;
	vt.mOriginY = documentBounds.top;
	vt.mPixelsPerMM = mmToInches * dpi;
	vt.mMMPerPixel = 1.0f / vt.mPixelsPerMM;
	Render(vt, 0, 0, w, h, true);

	// save as PNG
	VDPixmap px {};
	px.w = w;
	px.h = h;
	px.data = &mFrameBuffer[w * (h - 1)];
	px.pitch = -w*4;
	px.format = nsVDPixmap::kPixFormat_XRGB8888;

	ATSaveFrame(px, fn.c_str());

	// clear out large buffers
	vdfastvector<uint32> fb;
	fb.swap(mFrameBuffer);

	vdfastvector<uint8> ab;
	ab.swap(mABuffer);
}

void ATUIPrinterGraphicalOutputWindow::SaveAsPDF() {
	const VDStringW& fn = VDGetSaveFileName("PrinterSaveAsPDF"_vdtypeid, (VDGUIHandle)mhwnd, L"Save PDF", L"PDF document\0*.pdf\0", L"pdf");
	if (fn.empty())
		return;

	// ZLIB header (Deflate 32K Normal)
	static constexpr uint8 kZLIBHeader[] { 0x78, 0x9C };

	VDFileStream fileOut(fn.c_str(), nsVDFile::kWrite | nsVDFile::kCreateAlways | nsVDFile::kDenyAll);
	VDTextOutputStream textOut(&fileOut);

	vdfastvector<uint32> objectOffsets;

	textOut.PutLine("%PDF-1.4");
	textOut.PutLine("%\x80\x80\x80\x80");

	objectOffsets.push_back((uint32)textOut.Pos());
	textOut.PutLine("1 0 obj");
	textOut.PutLine("<< /Type /Catalog");
	textOut.PutLine("/Pages 2 0 R");
	textOut.PutLine(">>");
	textOut.PutLine("endobj");

	// reserve entry for pages object
	objectOffsets.push_back(0);

	// build TrueType font
	const auto& spec = mpOutput->GetGraphicsSpec();
	const int ttfDotsPerLine = std::min<int>(spec.mNumPins, 7);
	const float ttfAscentMM = spec.mDotSizeMM*2 + spec.mVerticalDotPitchMM * (ttfDotsPerLine - 1);
	const float ttfUnitsPerMM = 1000.0f / ttfAscentMM;
	const int ttfDotRadius = (int)(0.5f + mDotSizeMM * ttfUnitsPerMM);
	const int ttfDotVRange = 1000 - 2*ttfDotRadius;

	vdautoptr<ATTrueTypeEncoder> ttf(new ATTrueTypeEncoder);

	ttf->SetDefaultAdvanceWidth(ttfDotRadius * 2);

	// Create missing char
	ttf->BeginSimpleGlyph();
	ttf->EndSimpleGlyph();

	// Create space
	ttf->MapCharacter(0x20, ttf->BeginSimpleGlyph());
	ttf->EndSimpleGlyph();

	// Create initial dot
	//
	// TrueType uses quadratic splines with clockwise orientation in a bottom-up coordinate
	// system. We use 8 on-curve points at octants and round it off with another 8 off-curve
	// points. For a unit circle, the length of an octagonal segment is:
	//
	//  hypot(1-sqrt(2)/2, sqrt(2)/2) = sqrt(2 - sqrt(2)) = 0.765367
	//
	// The law of cosines then gives the distance for the control point:
	//
	//  c^2 = a^2 + b^2 - 2ab cos theta    (theta = 22.5d, a=b)
	//  2 - sqrt(2) = 2*a^2 * (1 - cos(135d))
	//  2 - sqrt(2) = 2*a^2 * (1 + sqrt(2)/2)
	//  a^2 = [2 - sqrt(2)] / [2 + sqrt(2)]
	//  a = sqrt(2) - 1

	auto dotGlyph = ttf->BeginSimpleGlyph();
	ttf->MapCharacter(0x21, dotGlyph);

	{
		const int r = ttfDotRadius;
		const int r2 = r*2;
		const int k = (ttfDotRadius * 724) / 1024;	// sqrt(2)/2
		const int c = (ttfDotRadius * 424) / 1024;	// sqrt(2) - 1

		ttf->AddGlyphPoint(r,   0,   true );
		ttf->AddGlyphPoint(r-c, 0,   false);
		ttf->AddGlyphPoint(r-k, r-k, true );
		ttf->AddGlyphPoint(0,   r-c, false);
		ttf->AddGlyphPoint(0,   r,   true );
		ttf->AddGlyphPoint(0,   r+c, false);
		ttf->AddGlyphPoint(r-k, r+k, true );
		ttf->AddGlyphPoint(r-c, r2,  false);
		ttf->AddGlyphPoint(r,   r2,  true );
		ttf->AddGlyphPoint(r+c, r2,  false);
		ttf->AddGlyphPoint(r+k, r+k, true );
		ttf->AddGlyphPoint(r2,  r+c, false);
		ttf->AddGlyphPoint(r2,  r,   true );
		ttf->AddGlyphPoint(r2,  r-c, false);
		ttf->AddGlyphPoint(r+k, r-k, true );
		ttf->AddGlyphPoint(r+c, 0,   false);
		ttf->EndContour();
	}

	ttf->EndSimpleGlyph();

	// map 0x20-0x7F
	for(int i=2; i<96; ++i) {
		ttf->MapCharacter(0x20+i, ttf->BeginCompositeGlyph());

		for(int j=0; j<7; ++j) {
			if (i & (1<<j))
				ttf->AddGlyphReference(dotGlyph, 0, (ttfDotVRange*j) / 6);
		}
		ttf->EndCompositeGlyph();
	}

	// Map 0xC0-0xDF. We avoid 0x80-0xBF because of character mapping problems in 0x80-0x9F and the altspace
	// at 0xA0. PDF is supposed to allow symbolic fonts to be mapped 1:1, but this doesn't seem to work in Acrobat
	// despite working in the browser viewers, and it's difficult to satisfy both PDF's symbol font character
	// mappings and FontValidator's requirements. Therefore we avoid the issue and just map to Latin-1 characters.
	for(int i=96; i<128; ++i) {
		ttf->MapCharacter(0x60+i, ttf->BeginCompositeGlyph());

		for(int j=0; j<7; ++j) {
			if (i & (1<<j))
				ttf->AddGlyphReference(dotGlyph, 0, (ttfDotVRange*j) / 6);
		}
		ttf->EndCompositeGlyph();
	}

	ttf->SetName(ATTrueTypeName::Copyright, "None - autogenerated");
	ttf->SetName(ATTrueTypeName::FontFamily, "Altirra Print");
	ttf->SetName(ATTrueTypeName::FontSubfamily, "Normal");
	ttf->SetName(ATTrueTypeName::FullFontName, "Altirra Print Normal");
	ttf->SetName(ATTrueTypeName::UniqueFontIdentifier, "Altirra Print Normal");
	ttf->SetName(ATTrueTypeName::Version, "Version 1.0");

	const vdspan<const uint8> fontData = ttf->Finalize();

	// add entry for embedded font (3)
	objectOffsets.push_back((uint32)textOut.Pos());
	textOut.FormatLine("%u 0 obj", (uint32)objectOffsets.size());

	textOut.PutLine(
		"<< /Type /Font "
		"/Subtype /TrueType "
		"/BaseFont /AAAAAA+Print "
		"/FirstChar 32 "
		"/LastChar 32 "
		"/Widths ["
	);

	textOut.FormatLine(" %d", 2*ttfDotRadius);

	textOut.PutLine(
		"] "
		"/FontDescriptor 4 0 R >>");

	textOut.PutLine("endobj");

	// add entry for embedded font descriptor (4)
	objectOffsets.push_back((uint32)textOut.Pos());
	textOut.FormatLine("%u 0 obj", (uint32)objectOffsets.size());

	textOut.FormatLine(
		"<< /Type /FontDescriptor "
		"/FontName /AAAAAA+Print "
		"/Flags 5 "
		"/FontBBox [0 -24 %d 1000] "
		"/ItalicAngle 0 "
		"/Ascent 1000 "
		"/Descent -24 "
		"/CapHeight 1000 "
		"/StemV 80 "
		"/MissingWidth %d "
		"/FontFile2 5 0 R >>",
		2 * ttfDotRadius,
		2 * ttfDotRadius
	);

	textOut.PutLine("endobj");

	// add entry for embedded font (5)
	objectOffsets.push_back((uint32)textOut.Pos());

	textOut.FormatLine("%u 0 obj", (uint32)objectOffsets.size());

	{
		VDMemoryBufferStream bufs;

		bufs.Write(kZLIBHeader, 2);

		uint32 adler32;

		{
			VDDeflateStream defs(bufs, VDDeflateChecksumMode::Adler32);
			defs.SetCompressionLevel(VDDeflateCompressionLevel::Quick);
			defs.Write(fontData.data(), fontData.size());
			defs.Finalize();
			adler32 = defs.Adler32();
		}

		uint32 v = VDToBEU32(adler32);
		bufs.Write(&v, 4);

		unsigned clen = (unsigned)bufs.Length();
		textOut.FormatLine("<</Length %u/Length1 %u/Filter/FlateDecode>>", clen, (unsigned)fontData.size());

		textOut.PutLine("stream");

		textOut.Write((const char *)bufs.GetBuffer().data(), clen);

		textOut.PutLine();
		textOut.PutLine("endstream");
	}

	textOut.PutLine("endobj");

	// render pages
	uint32 basePageObj = (uint32)objectOffsets.size() + 1;

	static constexpr float mmToPoints = 72.0f / 25.4f;
	const float pageWidthMM = mPageWidthMM;
	const float pageHeightMM = 11.0f * 25.4f;
	const float headHeightMM = spec.mDotSizeMM * 2 + spec.mVerticalDotPitchMM * (float)(spec.mNumPins - 1);
	const int numPages = std::max(1, (int)ceilf(mpOutput->GetDocumentBounds().bottom / pageHeightMM));

	const float lineToBaselineAdjustMM = spec.mbBit0Top ? headHeightMM - spec.mDotSizeMM : spec.mDotSizeMM;

	for(int page = 0; page < numPages; ++page) {
		objectOffsets.push_back((uint32)textOut.Pos());
		textOut.FormatLine("%u 0 obj", (uint32)objectOffsets.size());
		textOut.PutLine("<< /Type /Page");
		textOut.PutLine("/Parent 2 0 R");
		textOut.PutLine("/Resources <<");
		textOut.PutLine(" /Font << /Print 3 0 R >>");
		textOut.PutLine(">>");
		textOut.FormatLine("/Contents %u 0 R", (uint32)objectOffsets.size() + 1);
		textOut.PutLine(">>");
		textOut.PutLine("endobj");

		objectOffsets.push_back((uint32)textOut.Pos());
		textOut.FormatLine("%u 0 obj", (uint32)objectOffsets.size());

		VDStringA s;
		
		s.sprintf("%.3f g", powf(kBlackLevel, 1.0f/2.2f));

		if (mpOutput) {
			ATPrinterGraphicalOutput::CullInfo cullInfo {};
			vdrect32f pageRect(0.0f, pageHeightMM * (float)page, pageWidthMM, pageHeightMM * (float)(page + 1));

			if (mpOutput->PreCull(cullInfo, pageRect)) {
				const float mmToUnits = 10000.0f / 25.4f;

				vdfastvector<ATPrinterGraphicalOutput::RenderColumn> cols;
				float lineY = 0;
				while(mpOutput->ExtractNextLine(cols, lineY, cullInfo, pageRect)) {
					// if the head orientation is top-down, reverse the bit pattern to bottom-up
					if (spec.mbBit0Top) {
						int shift = 32 - spec.mNumPins;
						for(ATPrinterGraphicalOutput::RenderColumn& col : cols) {
							uint32 v = col.mPins;

							v = ((v & 0x55555555) << 1) + ((v >> 1) & 0x55555555);
							v = ((v & 0x33333333) << 2) + ((v >> 2) & 0x33333333);
							v = ((v & 0x0F0F0F0F) << 4) + ((v >> 4) & 0x0F0F0F0F);
							v = ((v & 0x00FF00FF) << 8) + ((v >> 8) & 0x00FF00FF);
							v = (v >> 16) + (v << 16);

							col.mPins = v >> shift;
						}
					}

					// encode 10000 units / 1" per tile
					const float unitsPerPoint = 10000.0f / 72.0f;
					const float pointsPerUnit = 1.0f / unitsPerPoint;

					// PDF by default scales the font's em size to 1 unit high, so
					// we need to scale by the desired height of the em square -- which
					// is 1.024 times the height in 10000th of an inch.
					const float fontSize = ttfAscentMM * mmToUnits * 1024.0f / 1000.0f;

					s.append_sprintf(
						" q %.10f 0 0 %.10f 0 0 cm /Print %.2f Tf"
						, pointsPerUnit
						, pointsPerUnit
						, fontSize
					);

					// The font only accommodates up to 7 pins, so if there are more than 7
					// pins in the head, we may need to do more than one band.
					const int numBands = (spec.mNumPins + 6) / 7;

					for(int band = 0; band < numBands; ++band) {
						// Find the first column that has anything in the band. We do this to detect if a band
						// is empty so we can skip the band.
						const int bandPinShift = 7 * band;
						vdspan bandColumns(cols);
						uint32 bandPins = 0x7F << bandPinShift;

						while(!bandColumns.empty()) {
							if (bandColumns.front().mPins & bandPins)
								break;

							bandColumns = bandColumns.subspan(1);
						}

						if (bandColumns.empty())
							continue;

						// Set the line origin and begin the text object. X is easy as it's just the left
						// edge, but Y needs to be adjusted. In PDF, it needs to be set to the baseline,
						// but it's the center of the first dot in the printer output.
						const int fxx0 = VDRoundToInt32((cols[0].mX - pageRect.left) * mmToUnits);
						const int fxy0 = VDRoundToInt32((pageRect.bottom - (lineY + lineToBaselineAdjustMM - spec.mVerticalDotPitchMM * bandPinShift)) * mmToUnits);

						// Sort the columns in the font by ascending X position so we have the smallest
						// delta X offsets.
						std::sort(cols.begin(), cols.end(),
							[](const auto& a, const auto& b) {
								return a.mX < b.mX;
							}
						);

						// begin text object, update text transform, and begin array for TJ command
						s.append_sprintf(" BT %d %d Td [", fxx0, fxy0);

						const float dotWidthMils = 2 * ttfDotRadius;
						const float mmToMils = 1000.0f * mmToUnits / fontSize;
						float xoff = (-pageRect.left * mmToUnits - fxx0) * 1000.0f / fontSize;
						for(const auto& col : bandColumns) {
							uint32 pins = (col.mPins >> bandPinShift) & 0x7F;

							if (pins) {
								const int dx = VDRoundToInt32(col.mX * mmToMils + xoff);

								// apply vertical offset if needed
								if (dx)
									s.append_sprintf("%d", -dx);

								// print pins using character
								s.append_sprintf("<%02X>", pins >= 0x60 ? 0x60 + pins : 0x20 + pins);

								// update X offset tracking based on advance width and applied adjustment
								xoff -= dotWidthMils + (float)dx;
							}
						}

						// print text and end text object
						s += "] TJ ET";
					}

					// end of tile - pop transform
					s += " Q";
				}
			}
		}

		VDMemoryBufferStream bufs;

		bufs.Write(kZLIBHeader, 2);

		uint32 adler32;

		{
			VDDeflateStream defs(bufs, VDDeflateChecksumMode::Adler32);
			defs.SetCompressionLevel(VDDeflateCompressionLevel::Quick);
			defs.Write(s.data(), s.size());
			defs.Finalize();
			adler32 = defs.Adler32();
		}

		uint32 v = VDToBEU32(adler32);
		bufs.Write(&v, 4);

		textOut.FormatLine("<< /Length %u /Filter /FlateDecode >>", (unsigned)bufs.Length());
		textOut.PutLine("stream");

		const auto bufData = bufs.GetBuffer();
		textOut.Write((const char *)bufData.data(), (int)bufData.size());

		textOut.PutLine();
		textOut.PutLine("endstream");
		textOut.PutLine("endobj");
	}

	// write pages table
	objectOffsets[1] = (uint32)textOut.Pos();
	textOut.PutLine("2 0 obj");
	textOut.PutLine("<< /Type /Pages");
	textOut.PutLine("/Kids [");

	for(int i=0; i<numPages; ++i)
		textOut.FormatLine("%u 0 R", basePageObj + i*2);

	textOut.PutLine("]");
	textOut.FormatLine("/Count %u", numPages);
	textOut.FormatLine("/MediaBox [0 0 %f %f]", mPageWidthMM * mmToPoints, pageHeightMM * mmToPoints);
	textOut.PutLine(">>");
	textOut.PutLine("endobj");

	// write cross-reference table
	uint32 xrefPos = (uint32)textOut.Pos();

	textOut.PutLine("xref");
	textOut.FormatLine("0 %u", (unsigned)objectOffsets.size() + 1);
	textOut.PutLine("0000000000 65535 f");

	for(uint32 offset : objectOffsets)
		textOut.FormatLine("%010u 00000 n", offset);

	textOut.PutLine("trailer");
	textOut.PutLine("<< /Root 1 0 R");
	textOut.FormatLine("/Size %u", (unsigned)(objectOffsets.size() + 1));
	textOut.PutLine(">> startxref");
	textOut.FormatLine("%u", xrefPos);
	textOut.PutLine("%%EOF");
}

LRESULT ATUIPrinterGraphicalOutputWindow::WndProc(UINT msg, WPARAM wParam, LPARAM lParam) {
	switch(msg) {
		case WM_CREATE:
			mCanvas.Init(mhwnd);
			SetTouchMode(kATUITouchMode_2DPanSmooth);
			break;

		case WM_DESTROY:
			OnDestroy();
			break;

		case WM_SIZE:
			OnSize();
			return 0;

		case WM_NCCALCSIZE:
			return OnNCCalcSize(wParam, lParam);

		case WM_KEYDOWN:
			OnKeyDown(wParam, lParam);
			return 0;

		case WM_MOUSEMOVE:
			OnMouseMove(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			return 0;

		case WM_LBUTTONDOWN:
		case WM_LBUTTONDBLCLK:
			Focus();
			OnMouseDownL(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			return 0;

		case WM_LBUTTONUP:
			OnMouseUpL(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			return 0;

		case WM_MOUSEWHEEL:
			OnMouseWheel(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam), (float)GET_WHEEL_DELTA_WPARAM(wParam) / (float)WHEEL_DELTA);
			return 0;

		case WM_CAPTURECHANGED:
			OnCaptureChanged();
			return 0;

		case WM_PAINT:
			OnPaint();
			return 0;

		case WM_ERASEBKGND:
			return 0;

		case WM_GESTURE:
			if (OnGesture(wParam, lParam))
				return 0;
			break;

		case WM_DPICHANGED:
		case ATWM_INHERIT_DPICHANGED:
			OnDpiChanged(wParam);
			break;

		case WM_USER + 100:
			ProcessPendingInvalidation();
			return 0;
	}

	return ATUINativeWindow::WndProc(msg, wParam, lParam);
}

LRESULT ATUIPrinterGraphicalOutputWindow::OnNCCalcSize(WPARAM wParam, LPARAM lParam) {
	if (!wParam)
		return DefWindowProc(mhwnd, WM_NCCALCSIZE, wParam, lParam);

	NCCALCSIZE_PARAMS& params = *(NCCALCSIZE_PARAMS *)lParam;

	// ask DWP to compute the new client size
	DefWindowProc(mhwnd, WM_NCCALCSIZE, FALSE, (LPARAM)&params.rgrc[0]);

	// compute the source and destination client rect sizes
	sint32 neww = params.rgrc[0].right  - params.rgrc[0].left;
	sint32 newh = params.rgrc[0].bottom - params.rgrc[0].top;
	sint32 oldw = params.rgrc[2].right  - params.rgrc[2].left;
	sint32 oldh = params.rgrc[2].bottom - params.rgrc[2].top;

	// compute the difference in center offsets
	sint32 newcx = neww / 2;
	sint32 newcy = newh / 2;
	sint32 oldcx = oldw / 2;
	sint32 oldcy = oldh / 2;

	// compute the size to copy
	sint32 savecx = std::min<sint32>(oldcx, newcx);
	sint32 savecy = std::min<sint32>(oldcy, newcy);
	sint32 savew = std::min<sint32>(oldw, neww);
	sint32 saveh = std::min<sint32>(oldh, newh);

	// set source and dest copy rects
	params.rgrc[1].left   = params.rgrc[0].left + newcx - savecx;
	params.rgrc[1].top    = params.rgrc[0].top  + newcy - savecy;
	params.rgrc[1].right  = params.rgrc[1].left + savew;
	params.rgrc[1].bottom = params.rgrc[1].top  + saveh;

	params.rgrc[2].left   = params.rgrc[2].left + oldcx - savecx;
	params.rgrc[2].top    = params.rgrc[2].top  + oldcy - savecy;
	params.rgrc[2].right  = params.rgrc[2].left + savew;
	params.rgrc[2].bottom = params.rgrc[2].top  + saveh;

	return WVR_VALIDRECTS;
}

void ATUIPrinterGraphicalOutputWindow::OnDestroy() {
	if (mpOutput) {
		mpOutput->SetOnInvalidation(nullptr);
		mpOutput = nullptr;
	}
}

void ATUIPrinterGraphicalOutputWindow::OnSize() {
	auto [w, h] = GetClientSize();

	if (mViewWidthPixels != w || mViewHeightPixels != h) {
		mViewWidthPixels = w;
		mViewHeightPixels = h;

		// Update the view origin based on the center and new size. We don't redraw
		// here as that's already done by the NCCALCSIZE correction.
		UpdateViewOrigin();
	}
}

void ATUIPrinterGraphicalOutputWindow::OnKeyDown(uint32 vk, uint32 flags) {
	const bool ctrl = GetKeyState(VK_CONTROL) < 0;

	switch(vk) {
		case VK_ESCAPE:
			if (ATGetUIPane(kATUIPaneId_Display))
				ATActivateUIPane(kATUIPaneId_Display, true);
			break;

		case VK_LEFT:
			ScrollByPixels(ctrl ? 1 : 100, 0);
			break;

		case VK_RIGHT:
			ScrollByPixels(ctrl ? -1 : -100, 0);
			break;

		case VK_UP:
			ScrollByPixels(0, ctrl ? 1 : 100);
			break;

		case VK_DOWN:
			ScrollByPixels(0, ctrl ? -1 : -100);
			break;

		case VK_PRIOR:
			ScrollByPixels(0, mViewHeightPixels);
			break;

		case VK_NEXT:
			ScrollByPixels(0, -mViewHeightPixels);
			break;

		case VK_OEM_PLUS:
			if (ctrl)
				SetZoom(mZoomClicks + 1.0f);
			break;

		case VK_OEM_MINUS:
			if (ctrl)
				SetZoom(mZoomClicks - 1.0f);
			break;
	}
}

void ATUIPrinterGraphicalOutputWindow::OnMouseMove(int x, int y) {
	if (mbDragging) {
		int dx = x - mDragLastX;
		int dy = y - mDragLastY;
		mDragLastX = x;
		mDragLastY = y;

		ScrollByPixels(dx, dy);
	}
}

void ATUIPrinterGraphicalOutputWindow::OnMouseDownL(int x, int y) {
	mbDragging = true;
	mDragLastX = x;
	mDragLastY = y;
	::SetCapture(mhwnd);
}

void ATUIPrinterGraphicalOutputWindow::OnMouseUpL(int x, int y) {
	if (mbDragging) {
		mbDragging = false;
		::ReleaseCapture();
	}
}

void ATUIPrinterGraphicalOutputWindow::OnMouseWheel(int x, int y, float delta) {
	mWheelAccum += delta;

	const int clicks = (int)floorf(mWheelAccum + 0.5f);

	if (clicks) {
		const vdpoint32& cpt = TransformScreenToClient(vdpoint32(x, y));
		SetZoom(mZoomClicks + (float)clicks, cpt);

		mWheelAccum -= (int)clicks;
	}
}

void ATUIPrinterGraphicalOutputWindow::OnCaptureChanged() {
	mbDragging = false;
}

bool ATUIPrinterGraphicalOutputWindow::OnGesture(WPARAM wParam, LPARAM lParam) {
	GESTUREINFO gestureInfo {sizeof(GESTUREINFO)};

	const BOOL haveInfo = GetGestureInfo((HGESTUREINFO)lParam, &gestureInfo);
	if (!haveInfo)
		return false;

	CloseGestureInfoHandle((HGESTUREINFO)lParam);

	switch(gestureInfo.dwID) {
		case GID_BEGIN:
			mbInGesture = true;
			mbFirstGestureEvent = true;
			break;

		case GID_END:
			mbInGesture = false;
			mbFirstGestureEvent = false;
			break;

		default:
			if (mbInGesture) {
				vdpoint32 pt(gestureInfo.ptsLocation.x, gestureInfo.ptsLocation.y);

				switch(gestureInfo.dwID) {
					case GID_PAN:
						if (!mbFirstGestureEvent)
							ScrollByPixels(pt.x - mGestureOrigin.x, pt.y - mGestureOrigin.y);

						mGestureOrigin = pt;
						break;

					case GID_ZOOM:
						if (float distance = (float)gestureInfo.ullArguments; mbFirstGestureEvent) {
							mGestureZoomOrigin = distance;
						} else {
							if (distance > mGestureZoomOrigin * 1.10f) {
								mGestureZoomOrigin *= 1.1487f;
								SetZoom(mZoomClicks + 1.0f);
							} else if (distance < mGestureZoomOrigin / 1.10f) {
								mGestureZoomOrigin /= 1.1487f;
								SetZoom(mZoomClicks - 1.0f);
							}
						}
						break;
				}

				mbFirstGestureEvent = false;
			}
			break;
	}

	return false;
}

void ATUIPrinterGraphicalOutputWindow::OnPaint() {
	PAINTSTRUCT ps;
	if (HDC hdc = mCanvas.BeginDirect(ps, false)) {
		const sint32 w = ps.rcPaint.right - ps.rcPaint.left;
		const sint32 h = ps.rcPaint.bottom - ps.rcPaint.top;

		if (w > 0 && h > 0) {
			RECT *paintRects = &ps.rcPaint;
			size_t numPaintRects = 1;

			HRGN hrgnUpdate = ::CreateRectRgn(0, 0, 0, 0);
			if (hrgnUpdate) {
				if (::GetRandomRgn(hdc, hrgnUpdate, SYSRGN) > 0) {
					DWORD numBytes = ::GetRegionData(hrgnUpdate, 0, nullptr);

					if (numBytes) {
						if (mRegionData.size() < numBytes)
							mRegionData.resize(numBytes);

						if (::GetRegionData(hrgnUpdate, mRegionData.size(), mRegionData.data())) {
							paintRects = (RECT *)mRegionData->Buffer;
							numPaintRects = mRegionData->rdh.nCount;

							// convert all rects from screen to client coords
							for(size_t i = 0; i < numPaintRects; ++i) {
								RECT& r = paintRects[i];

								const auto& cr = TransformScreenToClient(vdrect32(r.left, r.top, r.right, r.bottom));

								r.left = cr.left;
								r.top = cr.top;
								r.right = cr.right;
								r.bottom = cr.bottom;
							}
						}
					}
				}
				::DeleteObject(hrgnUpdate);
			}

			while(numPaintRects--) {
				const RECT& paintRect = *paintRects++;

				// convert left and right to document coordinates
				const float viewfx1 = mViewOriginX + (float)paintRect.left * mViewMMPerPixel;
				const float viewfx2 = mViewOriginX + (float)paintRect.right * mViewMMPerPixel;

				// if we are full left/right, just clear to background color
				SetBkColor(hdc, 0x808080);
				SetBkMode(hdc, OPAQUE);

				if (viewfx2 <= 0 || viewfx1 >= mPageWidthMM) {
					ExtTextOutW(hdc, 0, 0, ETO_OPAQUE, &paintRect, L"", 0, nullptr);
				} else {
					// compute clipped page width within paint bounds, now that we know we won't
					// massively overflow
					sint32 paintx1 = paintRect.left;
					sint32 paintx2 = paintRect.right;
					sint32 painty1 = paintRect.top;
					sint32 painty2 = paintRect.bottom;

					if (viewfx1 < 0) {
						paintx1 = (sint32)ceilf(-mViewOriginX * mViewPixelsPerMM - 0.5f / 8.0f);

						RECT rLeft  = paintRect;
						rLeft.right = paintx1;

						ExtTextOutW(hdc, 0, 0, ETO_OPAQUE, &rLeft, L"", 0, nullptr);
					}

					if (viewfx2 > mPageWidthMM) {
						paintx2 = (sint32)ceilf((mPageWidthMM - mViewOriginX) * mViewPixelsPerMM - 7.5f / 8.0f);

						RECT rRight  = paintRect;
						rRight.left = paintx2;

						ExtTextOutW(hdc, 0, 0, ETO_OPAQUE, &rRight, L"", 0, nullptr);
					}

					// check for clipping at top of page
					const float viewfy1 = mViewOriginY + (float)paintRect.top * mViewMMPerPixel;
					if (viewfy1 < 0) {
						painty1 = (sint32)ceilf(-mViewOriginY * mViewPixelsPerMM - 0.5f / 8.0f);

						RECT rTop = paintRect;
						rTop.left = paintx1;
						rTop.right = paintx2;
						rTop.bottom = painty1;

						ExtTextOutW(hdc, 0, 0, ETO_OPAQUE, &rTop, L"", 0, nullptr);
					}

					const sint32 paintw = paintx2 - paintx1;
					const sint32 painth = painty2 - painty1;

					if (paintw > 0 && painth > 0) {
						ViewTransform viewTransform;
						viewTransform.mOriginX = mViewOriginX;
						viewTransform.mOriginY = mViewOriginY;
						viewTransform.mMMPerPixel = mViewMMPerPixel;
						viewTransform.mPixelsPerMM = mViewPixelsPerMM;

						if (Render(viewTransform, paintx1, painty1, paintw, painth, false)) {
							BITMAPINFO bi {};
							bi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
							bi.bmiHeader.biWidth = paintw;
							bi.bmiHeader.biHeight = painth;
							bi.bmiHeader.biPlanes = 1;
							bi.bmiHeader.biBitCount = 32;
							bi.bmiHeader.biCompression = BI_RGB;
							bi.bmiHeader.biSizeImage = paintw * h * 4;

							SetDIBitsToDevice(hdc, paintx1, painty1, paintw, painth, 0, 0, 0, painth, mFrameBuffer.data(), &bi, DIB_RGB_COLORS);
						} else {
							SetBkColor(hdc, VDSwizzleU32(mGammaTable[0]) >> 8);

							const RECT r { paintx1, painty1, paintx2, painty2 };

							ExtTextOutW(hdc, 0, 0, ETO_OPAQUE, &r, L"", 0, nullptr);
						}
					}
				}
			}
		}

		mCanvas.EndDirect(ps);
	}
}

void ATUIPrinterGraphicalOutputWindow::OnDpiChanged(int dpi) {
	if (dpi > 0 && dpi < 10000) {
		if (mViewDpi != dpi)
			UpdateZoom(nullptr);
	}
}

void ATUIPrinterGraphicalOutputWindow::UpdateViewOrigin() {
	mViewOriginX = mViewCenterX - (float)(mViewWidthPixels  / 2) * mViewMMPerPixel;
	mViewOriginY = mViewCenterY - (float)(mViewHeightPixels / 2) * mViewMMPerPixel;
}

void ATUIPrinterGraphicalOutputWindow::ScrollByPixels(int dx, int dy) {
	if (dx || dy) {
		mViewCenterX -= mViewMMPerPixel * dx;
		mViewCenterY -= mViewMMPerPixel * dy;
		UpdateViewOrigin();

		mCanvas.Scroll(dx, dy);
	}
}

void ATUIPrinterGraphicalOutputWindow::SetZoom(float clicks) {
	SetZoom(clicks, vdpoint32(mViewWidthPixels / 2, mViewHeightPixels / 2));
}

void ATUIPrinterGraphicalOutputWindow::SetZoom(float clicks, const vdpoint32& centerPt) {
	float newZoom = std::clamp(clicks, kZoomMin, kZoomMax);

	if (fabsf(mZoomClicks - newZoom) < 0.0001f)
		return;

	mZoomClicks = newZoom;

	UpdateZoom(&centerPt);
}

void ATUIPrinterGraphicalOutputWindow::UpdateZoom(const vdpoint32* centerPt) {
	float basePixelsPerMM = (float)mViewDpi / 25.4f;
	float newPixelsPerMM = basePixelsPerMM * powf(2.0f, (float)mZoomClicks / 5.0f);

	if (fabsf(mViewPixelsPerMM - newPixelsPerMM) > 1e-3f*std::max(mViewPixelsPerMM, newPixelsPerMM)) {
		const float pivotDX = centerPt ? centerPt->x - mViewWidthPixels / 2 : 0.0f;
		const float pivotDY = centerPt ? centerPt->y - mViewHeightPixels / 2 : 0.0f;

		float anchorX = mViewCenterX + mViewMMPerPixel * pivotDX;
		float anchorY = mViewCenterY + mViewMMPerPixel * pivotDY;

		mViewPixelsPerMM = newPixelsPerMM;
		mViewMMPerPixel = 1.0f / mViewPixelsPerMM;

		mViewCenterX = anchorX - mViewMMPerPixel * pivotDX;
		mViewCenterY = anchorY - mViewMMPerPixel * pivotDY;
		UpdateViewOrigin();

		ForceFullInvalidation();
	}
}

void ATUIPrinterGraphicalOutputWindow::ForceFullInvalidation() {
	InvalidateRect(mhwnd, nullptr, false);
	if (mpOutput) {
		bool all;
		vdrect32f r;
		mpOutput->ExtractInvalidationRect(all, r);
	}
}

void ATUIPrinterGraphicalOutputWindow::ProcessPendingInvalidation() {
	mbInvalidationPending = false;

	if (mpOutput) {
		bool invAll;
		vdrect32f invRect;

		if (mpOutput->ExtractInvalidationRect(invAll, invRect)) {
			if (invAll)
				InvalidateRect(mhwnd, nullptr, false);
			else
				InvalidatePaperArea(invRect.left, invRect.top, invRect.right, invRect.bottom);
		}
	}
}

void ATUIPrinterGraphicalOutputWindow::InvalidatePaperArea(float x1, float y1, float x2, float y2) {
	if (x1 >= x2 || y1 >= y2)
		return;

	x1 -= mViewOriginX;
	x2 -= mViewOriginX;
	y1 -= mViewOriginY;
	y2 -= mViewOriginY;

	x1 = x1 * mViewPixelsPerMM;
	y1 = y1 * mViewPixelsPerMM;
	x2 = x2 * mViewPixelsPerMM;
	y2 = y2 * mViewPixelsPerMM;

	if (x1 < 0)
		x1 = 0;

	if (y1 < 0)
		y1 = 0;

	if (x2 > (float)mViewWidthPixels)
		x2 = (float)mViewWidthPixels;

	if (y2 > (float)mViewHeightPixels)
		y2 = (float)mViewHeightPixels;

	if (x1 >= x2 || y1 >= y2)
		return;

	RECT r {
		(LONG)ceilf(x1 - 0.5f - 3.5f / 8.0f),
		(LONG)ceilf(y1 - 0.5f - 3.5f / 8.0f),
		(LONG)ceilf(x2 - 0.5f + 3.5f / 8.0f),
		(LONG)ceilf(y2 - 0.5f + 3.5f / 8.0f)
	};

	if (r.left < r.right && r.top < r.bottom)
		InvalidateRect(mhwnd, &r, FALSE);
}

vdrect32f ATUIPrinterGraphicalOutputWindow::GetDocumentBounds() const {
	if (mpOutput)
		return mpOutput->GetDocumentBounds();
	else
		return vdrect32f(0, 0, 0, 0);
}

bool ATUIPrinterGraphicalOutputWindow::Render(const ViewTransform& viewTransform, sint32 x, sint32 y, uint32 w, uint32 h, bool force) {
	// compute pixel bounding box
	const sint32 viewix1 = x;
	const sint32 viewiy1 = y;
	const sint32 viewix2 = x + w;
	const sint32 viewiy2 = y + h;
	const float viewfx1 = (float)viewix1;
	const float viewfy1 = (float)viewiy1;
	const float viewfx2 = (float)viewix2;
	const float viewfy2 = (float)viewiy2;

	// compute document space bounding box
	const float viewDocX1 = viewTransform.mOriginX + viewfx1 * viewTransform.mMMPerPixel;
	const float viewDocY1 = viewTransform.mOriginY + viewfy1 * viewTransform.mMMPerPixel;
	const float viewDocX2 = viewTransform.mOriginX + viewfx2 * viewTransform.mMMPerPixel;
	const float viewDocY2 = viewTransform.mOriginY + viewfy2 * viewTransform.mMMPerPixel;

	// pre-cull line range to entire vertical range
	ATPrinterGraphicalOutput::CullInfo cullInfo;

	if (mpOutput) {
		if (!mpOutput->PreCull(cullInfo, vdrect32f(viewDocX1, viewDocY1, viewDocX2, viewDocY2)) && !force)
			return false;
	}

	// clear framebuffer
	mFrameBuffer.resize(w * h);
	std::fill(mFrameBuffer.begin(), mFrameBuffer.end(), 0xFFFFFFFF);

	// initialize antialiasing buffer
	if (mABuffer.size() < w*8)
		mABuffer.resize(w*8);

	// render one scan line at a time
	const float viewSubPixelsPerMM = viewTransform.mPixelsPerMM * 8.0f;
	const float rowCenterToFirstSubRowOffset = viewTransform.mMMPerPixel * (-3.5f / 8.0f);
	const float subRowStep = viewTransform.mMMPerPixel / 8.0f;
	const float dotRadius = mDotSizeMM;
	const float dotRadiusSq = dotRadius * dotRadius;
	const sint32 subw = w * 8;

	for(uint32 yoff = 0; yoff < h; ++yoff) {
		const float docRowYC = viewDocY1 + ((float)yoff + 0.5f) * viewTransform.mMMPerPixel;
		const float docRowYD = 0.5f * viewTransform.mMMPerPixel;

		// pre-cull lines to scan line vertical range
		const float docRowY1 = docRowYC - docRowYD;
		const float docRowY2 = docRowYC + docRowYD;

		// pre-cull dots to scan line
		mDotCullBuffer.clear();
		if (mpOutput)
			mpOutput->ExtractNextLineDots(mDotCullBuffer, cullInfo, vdrect32f(viewDocX1, docRowY1, viewDocX2, docRowY2));

		// clear the antialiasing buffer
		uint8 *abuf = mABuffer.data();
		memset(abuf, 0, w * 8);

		// render all dots
		const float docRowYDPlusDotR = docRowYD + dotRadius;
		for(const RenderDot& dot : mDotCullBuffer) {
			float dy = dot.mY - docRowYC;

			// vertical cull to row
			if (fabsf(dy) >= docRowYDPlusDotR)
				continue;

			// offset vertical delta to first subrow
			dy += rowCenterToFirstSubRowOffset;

			// process one subrow at a time
			for(uint32 subRow = 0; subRow < 8; ++subRow, (dy += subRowStep)) {
				const float dx2 = dotRadiusSq - dy*dy;

				if (dx2 <= 0)
					continue;

				const float dx = sqrtf(dx2);

				// compute x range in document space
				const float xc = dot.mX - viewDocX1;
				const float x1 = xc - dx;
				const float x2 = xc + dx;

				// convert x range to subpixels
				const sint32 subx1 = (sint32)ceilf(x1 * viewSubPixelsPerMM - 0.5f);
				const sint32 subx2 = (sint32)ceilf(x2 * viewSubPixelsPerMM - 0.5f);

				// horizontally clip
				const sint32 csubx1 = std::max<sint32>(subx1, 0);
				const sint32 csubx2 = std::min<sint32>(subx2, subw);

				if (csubx1 >= csubx2)
					continue;

				// draw bits
				const uint32 ucsubx1 = (uint32)csubx1;
				const uint32 ucsubx2 = (uint32)csubx2;
				const uint32 maskx1 = ucsubx1 >> 3;
				const uint32 maskx2 = (ucsubx2 - 1) >> 3;
				const uint8 mask1 = 0xFF >> (ucsubx1 & 7);
				const uint8 mask2 = 0xFF << ((8 - ucsubx2) & 7);
				uint8 *__restrict dst = abuf + subRow + maskx1*8;

				if (maskx1 == maskx2) {
					*dst |= mask1 & mask2;
				} else {
					*dst |= mask1;
					dst += 8;

					for(uint32 x = maskx1 + 1; x < maskx2; ++x) {
						*dst = 0xFF;
						dst += 8;
					}

					*dst |= mask2;
				}
			}
		}

		// render antialiasing buffer row to framebuffer
		Downsample8x8(&mFrameBuffer[w * (h - 1 - yoff)], abuf, w);
	}

	return true;
}

void ATUIPrinterGraphicalOutputWindow::Downsample8x8(uint32 *dst, const uint8 *src, size_t w) {
#ifdef VD_CPU_X64
	if (VDCheckAllExtensionsEnabled(VDCPUF_SUPPORTS_POPCNT))
		return Downsample8x8_POPCNT64(dst, src, w);
#endif

#ifdef VD_CPU_X86
	if (VDCheckAllExtensionsEnabled(VDCPUF_SUPPORTS_POPCNT))
		return Downsample8x8_POPCNT32(dst, src, w);
#endif

	Downsample8x8_Scalar(dst, src, w);
}

void ATUIPrinterGraphicalOutputWindow::Downsample8x8_Scalar(uint32 *dst, const uint8 *src, size_t w) {
	uint32 *__restrict fbdst = dst;
	const uint8 *__restrict asrc = src;

	for(uint32 x = 0; x < w; ++x) {
		// compute coverage by counting bits within 8x8 window
		const uint32 level
			= mPopCnt8[asrc[0]]
			+ mPopCnt8[asrc[1]]
			+ mPopCnt8[asrc[2]]
			+ mPopCnt8[asrc[3]]
			+ mPopCnt8[asrc[4]]
			+ mPopCnt8[asrc[5]]
			+ mPopCnt8[asrc[6]]
			+ mPopCnt8[asrc[7]];
		asrc += 8;

		// convert coverage to sRGB color
		*fbdst++ = mGammaTable[level];
	}
}

#ifdef VD_CPU_X64
void ATUIPrinterGraphicalOutputWindow::Downsample8x8_POPCNT64(uint32 *dst, const uint8 *src, size_t w) {
	uint32 *__restrict fbdst = dst;
	const uint8 *__restrict asrc = src;

	for(uint32 x = 0; x < w; ++x) {
		// compute coverage by counting bits within 8x8 window
		const size_t level = _mm_popcnt_u64(VDReadUnalignedU64(asrc));
		asrc += 8;

		// convert coverage to sRGB color
		*fbdst++ = mGammaTable[level];
	}
}
#endif

#ifdef VD_CPU_X86
void ATUIPrinterGraphicalOutputWindow::Downsample8x8_POPCNT32(uint32 *dst, const uint8 *src, size_t w) {
	uint32 *__restrict fbdst = dst;
	const uint8 *__restrict asrc = src;

	for(uint32 x = 0; x < w; ++x) {
		// compute coverage by counting bits within 8x8 window
		const size_t level = _mm_popcnt_u32(VDReadUnalignedU32(asrc))
			+ _mm_popcnt_u32(VDReadUnalignedU32(asrc + 4));
		asrc += 8;

		// convert coverage to sRGB color
		*fbdst++ = mGammaTable[level];
	}
}
#endif

///////////////////////////////////////////////////////////////////////////////

ATPrinterOutputWindow::ATPrinterOutputWindow()
	: ATUIPaneWindow(kATUIPaneId_PrinterOutput, L"Printer Output")
	, mhwndTextEditor(NULL)
	, mLineBufIdx(0)
{
	mPreferredDockCode = kATContainerDockBottom;

	mAddedOutputFn = [this](ATPrinterOutput& output) { OnAddedOutput(output); };
	mRemovingOutputFn = [this](ATPrinterOutput& output) { OnRemovingOutput(output); };
	mAddedGraphicalOutputFn = [this](ATPrinterGraphicalOutput& output) { OnAddedGraphicalOutput(output); };
	mRemovingGraphicalOutputFn = [this](ATPrinterGraphicalOutput& output) { OnRemovingGraphicalOutput(output); };
}

ATPrinterOutputWindow::~ATPrinterOutputWindow() {
}

LRESULT ATPrinterOutputWindow::WndProc(UINT msg, WPARAM wParam, LPARAM lParam) {
	switch(msg) {
		case WM_SIZE:
			OnSize();
			break;

		case WM_CONTEXTMENU:
			try {
				int x = GET_X_LPARAM(lParam);
				int y = GET_Y_LPARAM(lParam);

				HMENU menu0 = LoadMenu(NULL, mpGraphicWindow ? MAKEINTRESOURCE(IDR_PRINTER_GRAPHIC_CONTEXT_MENU) : MAKEINTRESOURCE(IDR_PRINTER_CONTEXT_MENU));
				if (menu0) {
					HMENU menu = GetSubMenu(menu0, 0);
					BOOL cmd = 0;

					if (x == -1 && y == -1) {
						const vdpoint32& pt = mpGraphicWindow
							? mpGraphicWindow->TransformClientToScreen(vdpoint32(0, 0))
							: mpTextEditor->GetScreenPosForContextMenu();
						
						x = pt.x;
						y = pt.y;
					} else {
						POINT pt = {x, y};

						if (!mpGraphicWindow && ScreenToClient(mhwndTextEditor, &pt))
							mpTextEditor->SetCursorPixelPos(pt.x, pt.y);
					}

					cmd = TrackPopupMenu(menu, TPM_LEFTALIGN|TPM_TOPALIGN|TPM_RETURNCMD, x, y, 0, mhwnd, NULL);

					DestroyMenu(menu0);

					switch(cmd) {
						case ID_CONTEXT_CLEAR:
							if (mpGraphicWindow)
								mpGraphicWindow->Clear();
							else if (mpTextEditor) {
								mpTextEditor->Clear();
								mLastTextOffset = 0;

								if (mpTextOutput) {
									mpTextOutput->Clear();
									mpTextOutput->Revalidate();
								}
							}
							break;

						case ID_CONTEXT_RESETVIEW:
							if (mpGraphicWindow)
								mpGraphicWindow->ResetView();
							break;

						case ID_SAVEAS_PNGIMAGE96:
							if (mpGraphicWindow)
								mpGraphicWindow->SaveAsPNG(96.0f);
							break;

						case ID_SAVEAS_PNGIMAGE300:
							if (mpGraphicWindow)
								mpGraphicWindow->SaveAsPNG(300.0f);
							break;

						case ID_SAVEAS_PDF:
							if (mpGraphicWindow)
								mpGraphicWindow->SaveAsPDF();
							break;
					}
				}
			} catch(const VDException& ex) {
				VDDialogFrameW32::ShowError((VDGUIHandle)mhwnd, ex.wc_str(), nullptr);
			}
			break;
	}

	return ATUIPaneWindow::WndProc(msg, wParam, lParam);
}

bool ATPrinterOutputWindow::OnCreate() {
	if (!ATUIPaneWindow::OnCreate())
		return false;

	if (!VDCreateTextEditor(~mpTextEditor))
		return false;

	mhwndTextEditor = (HWND)mpTextEditor->Create(WS_EX_NOPARENTNOTIFY, WS_CHILD|WS_VISIBLE, 0, 0, 0, 0, (VDGUIHandle)mhwnd, 100);

	OnFontsUpdated();

	mpTextEditor->SetReadOnly(true);

	OnSize();

	mpOutputMgr = static_cast<ATPrinterOutputManager *>(&g_sim.GetPrinterOutputManager());
	mpOutputMgr->OnAddedOutput.Add(&mAddedOutputFn);
	mpOutputMgr->OnRemovingOutput.Add(&mRemovingOutputFn);
	mpOutputMgr->OnAddedGraphicalOutput.Add(&mAddedGraphicalOutputFn);
	mpOutputMgr->OnRemovingGraphicalOutput.Add(&mRemovingGraphicalOutputFn);

	AttachToAnyOutput();

	return true;
}

void ATPrinterOutputWindow::OnDestroy() {
	DetachFromTextOutput();
	DetachFromGraphicsOutput();

	if (mpOutputMgr) {
		mpOutputMgr->OnAddedOutput.Remove(&mAddedOutputFn);
		mpOutputMgr->OnRemovingOutput.Remove(&mRemovingOutputFn);
		mpOutputMgr->OnAddedGraphicalOutput.Remove(&mAddedGraphicalOutputFn);
		mpOutputMgr->OnRemovingGraphicalOutput.Remove(&mRemovingGraphicalOutputFn);
		mpOutputMgr = nullptr;
	}

	ATUIPaneWindow::OnDestroy();
}

void ATPrinterOutputWindow::OnSize() {
	if (mpGraphicWindow) {
		mpGraphicWindow->SetArea(GetClientArea());
	} else if (mhwndTextEditor) {
		ATUINativeWindowProxy proxy(mhwndTextEditor);

		proxy.SetArea(GetClientArea());
	}
}

void ATPrinterOutputWindow::OnFontsUpdated() {
	if (mhwndTextEditor)
		SendMessage(mhwndTextEditor, WM_SETFONT, (WPARAM)ATGetConsoleFontW32(), TRUE);
}

void ATPrinterOutputWindow::OnSetFocus() {
	if (mpGraphicWindow)
		mpGraphicWindow->Focus();
	else
		::SetFocus(mhwndTextEditor);
}

void ATPrinterOutputWindow::OnAddedOutput(ATPrinterOutput& output) {
	if (!mpTextOutput && !mpGraphicsOutput)
		AttachToAnyOutput();
}

void ATPrinterOutputWindow::OnRemovingOutput(ATPrinterOutput& output) {
	if (&output == mpTextOutput) {
		DetachFromTextOutput();

		AttachToAnyOutput();
	}
}

void ATPrinterOutputWindow::OnAddedGraphicalOutput(ATPrinterGraphicalOutput& output) {
	if (!mpGraphicsOutput)
		AttachToAnyOutput();
}

void ATPrinterOutputWindow::OnRemovingGraphicalOutput(ATPrinterGraphicalOutput& output) {
	if (&output == mpGraphicsOutput) {
		DetachFromGraphicsOutput();

		AttachToAnyOutput();
	}
}

void ATPrinterOutputWindow::AttachToAnyOutput() {
	if (mpOutputMgr) {
		if (mpOutputMgr->GetGraphicalOutputCount() > 0)
			AttachToGraphicsOutput(mpOutputMgr->GetGraphicalOutput(0));
		else if (mpOutputMgr->GetOutputCount() > 0)
			AttachToTextOutput(mpOutputMgr->GetOutput(0));
	}
}

void ATPrinterOutputWindow::AttachToTextOutput(ATPrinterOutput& output) {
	DetachFromGraphicsOutput();

	mLastTextOffset = 0;
	if (mpTextEditor)
		mpTextEditor->Clear();

	mpTextOutput = &output;
	mpTextOutput->SetOnInvalidation(
		[this] {
			UpdateTextOutput();
		}
	);

	UpdateTextOutput();
}

void ATPrinterOutputWindow::DetachFromTextOutput() {
	if (mpTextOutput) {
		mpTextOutput->SetOnInvalidation(nullptr);
		mpTextOutput = nullptr;

		if (mpTextEditor)
			mpTextEditor->Clear();
	}
}

void ATPrinterOutputWindow::UpdateTextOutput() {
	if (mpTextOutput) {
		const size_t offset = mpTextOutput->GetLength();

		if (offset > mLastTextOffset) {
			if (mpTextEditor)
				mpTextEditor->Append(mpTextOutput->GetTextPointer(mLastTextOffset));

			mLastTextOffset = offset;
		}

		mpTextOutput->Revalidate();
	}
}

void ATPrinterOutputWindow::AttachToGraphicsOutput(ATPrinterGraphicalOutput& output) {
	DetachFromTextOutput();
	DetachFromGraphicsOutput();

	mpGraphicsOutput = &output;

	mpTextEditor->Clear();

	ATUINativeWindowProxy proxy(mhwndTextEditor);
	proxy.Hide();

	mpGraphicWindow = new ATUIPrinterGraphicalOutputWindow;
	mpGraphicWindow->CreateChild(mhdlg, 101, 0, 0, 0, 0, WS_CHILD | WS_VISIBLE | WS_TABSTOP);

	OnSize();

	mpGraphicWindow->AttachToOutput(output);
}

void ATPrinterOutputWindow::DetachFromGraphicsOutput() {
	if (mpGraphicsOutput) {
		mpGraphicsOutput->SetOnInvalidation(nullptr);
		mpGraphicsOutput = nullptr;
	}

	if (mpGraphicWindow) {
		mpGraphicWindow->Destroy();
		mpGraphicWindow = nullptr;

		// the text window was not resized while the graphic window existed, so fix that now
		OnSize();

		mpTextEditor->Clear();

		ATUINativeWindowProxy proxy(mhwndTextEditor);
		proxy.Show();
	}
}

////////////////////////////////////////////////////////////////////////////////

void ATUIDebuggerRegisterPrinterOutputPane() {
	ATRegisterUIPaneType(kATUIPaneId_PrinterOutput, VDRefCountObjectFactory<ATPrinterOutputWindow, ATUIPane>);
}
